import troposphere.elasticloadbalancing as elb
import troposphere.elasticloadbalancingv2 as elbv2
import troposphere.ec2 as ec2

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)
from .ec2 import SecurityGroupIngressInstanceELBPorts


# Temporary fix for https://github.com/cloudtools/troposphere/issues/1474
elbv2.one_of = my_one_of


# S - V2 LOAD BALANCING #
class ELBV2ListernerRuleECS(elbv2.ListenerRule):
    def __init__(self, title, key, index, mapname, scheme, **kwargs):
        super().__init__(title, **kwargs)

        auto_get_props(self, f"ListenerRules{index}")
        if "Conditions" not in key:
            self.Conditions = []
            if "HostHeader" in key:
                self.Conditions.append(
                    elbv2.Condition(
                        Field="host-header",
                        Values=[get_endvalue(f"{mapname}HostHeader", issub=True)],
                    )
                )
            if "PathPattern" in key:
                self.Conditions.append(
                    elbv2.Condition(
                        Field="path-pattern",
                        Values=[
                            Sub(
                                "${Value}",
                                **{"Value": get_endvalue(f"{mapname}PathPattern")},
                            )
                        ],
                    )
                )
            if "RequestMethod" in key:
                self.Conditions.append(
                    elbv2.Condition(
                        Field="http-request-method",
                        HttpRequestMethodConfig=elbv2.HttpRequestMethodConfig(
                            Values=[get_endvalue(f"{mapname}RequestMethod", issub=True)]
                        ),
                    )
                )
        if "Actions" not in key:
            tg_prefix = key.get("TargetGroupPrefix", "")
            self.Actions = [
                elbv2.Action(
                    Type="forward",
                    TargetGroupArn=Ref(f"TargetGroup{tg_prefix}{scheme}"),
                )
            ]
        self.Priority = get_endvalue(f"{mapname}Priority")


# E - V2 LOAD BALANCING #


def enable_recordset(rtype):
    prefix = f"Route53RecordSet"
    if "External" in cfg.RecordSet:
        if "External" in cfg.LoadBalancer:
            record = getattr(cfg, f"{prefix}{rtype}ExternalLoadBalancerExternal")
        else:
            record = getattr(cfg, f"{prefix}{rtype}ExternalLoadBalancerInternal")
        record["IBOX_ENABLED"] = True
    if "Internal" in cfg.RecordSet:
        if "Internal" in cfg.LoadBalancer:
            record = getattr(cfg, f"{prefix}{rtype}InternalLoadBalancerInternal")
        else:
            record = getattr(cfg, f"{prefix}{rtype}InternalLoadBalancerExternal")
        record["IBOX_ENABLED"] = True


def LB_ListenersEC2():
    # Resources
    Listeners = []
    for n, v in cfg.Listeners.items():
        mapname = f"Listeners{n}"  # Ex ListenersPort5601

        if cfg.LoadBalancerType == "Classic":
            Listener = elb.Listener(mapname)
            auto_get_props(Listener)
            auto_get_props(Listener, "ListenerClassic")
            Listeners.append(Listener)

        # resources
        r_SGIInstance = SecurityGroupIngressInstanceELBPorts(
            f"SecurityGroupIngress{mapname}", listener=mapname
        )

        if cfg.LoadBalancerType == "Network":
            r_SGIInstance.CidrIp = "0.0.0.0/0"
            del r_SGIInstance.properties["SourceSecurityGroupId"]
        else:
            # outputs
            Listener_Output = Output(mapname)
            Listener_Output.Value = Sub(
                "Protocol=${Protocol},Access=${Access}",
                **{
                    "Protocol": get_endvalue(f"{mapname}Protocol"),
                    "Access": get_endvalue(f"{mapname}Access"),
                },
            )
            add_obj(Listener_Output)

        add_obj(r_SGIInstance)

    return Listeners


def LB_ListenerRulesExternalInternal(index, key, mapname, scheme):
    # Skip if Rule is only for External/Internal
    only_scheme = key.get("Scheme")
    if only_scheme and only_scheme != scheme:
        return

    # resources
    R_RuleHttp = ELBV2ListernerRuleECS(
        f"ListenerHttp{scheme}Rules{index}",
        key=key,
        index=index,
        mapname=mapname,
        scheme=scheme,
    )
    R_RuleHttp.ListenerArn = get_expvalue(
        f"ListenerHttpDefault{scheme}", "LoadBalancerApplicationStack"
    )

    R_RuleHttps = ELBV2ListernerRuleECS(
        f"ListenerHttps{scheme}Rules{index}",
        key=key,
        index=index,
        mapname=mapname,
        scheme=scheme,
    )
    R_RuleHttps.ListenerArn = get_expvalue(
        f"ListenerHttpsDefault{scheme}", "LoadBalancerApplicationStack"
    )

    # Create ListenerRule only in stack's specific new Listener
    ListenerHttpPort = cfg.ListenerLoadBalancerHttpPort
    ListenerHttpsPort = cfg.ListenerLoadBalancerHttpsPort

    Protocol = key.get("Protocol", "auto")

    RuleHttpAdd = None
    RuleHttpsAdd = None

    if "NoDefault" in key:
        R_RuleHttp.ListenerArn = Ref(f"ListenerHttp{scheme}")
        R_RuleHttps.ListenerArn = Ref(f"ListenerHttps{scheme}")
        if ListenerHttpPort != 80:
            RuleHttpAdd = True
        if ListenerHttpsPort != 443:
            RuleHttpsAdd = True
    else:
        # by default create http rules on Internal LB and https rules on External one
        if scheme == "Internal":
            RuleHttpAdd = True
        if scheme == "External":
            RuleHttpsAdd = True
            # on External can be forced or overriden by key http/https/any
            if Protocol == "http":
                RuleHttpAdd = True
                RuleHttpsAdd = None
            if Protocol == "https":
                RuleHttpAdd = None
                RuleHttpsAdd = True
            if Protocol == "any":
                RuleHttpAdd = True
                RuleHttpsAdd = True

    if RuleHttpAdd:
        add_obj(R_RuleHttp)
    if RuleHttpsAdd:
        add_obj(R_RuleHttps)


def LB_ListenerRules():
    for n, v in cfg.ListenerRules.items():
        mapname = f"ListenerRules{n}"  # Ex. ListenerRules1

        # parameters
        p_Priority = Parameter(
            f"{mapname}Priority",
            Description="Listener Rule Priority, lesser value = high priority - empty for default based on env/role",
        )

        add_obj(p_Priority)

        ListenerRule_Out_String = ["Priority=${Priority}"]
        ListenerRule_Out_Map = {"Priority": get_endvalue(f"{mapname}Priority")}

        if "HostHeader" in v:
            p_HostHeader = Parameter(
                f"{mapname}HostHeader",
                Description="Listener Rule HostHeader Condition - empty for default based on env/role",
            )

            add_obj(p_HostHeader)

            # outputs
            ListenerRule_Out_String.append("HostHeader=${HostHeader}")
            ListenerRule_Out_Map.update(
                {"HostHeader": get_endvalue(f"{mapname}HostHeader", issub=True)}
            )

        if "PathPattern" in v:
            p_PathPattern = Parameter(
                f"{mapname}PathPattern",
                Description="Listener Rule PathPattern Condition - empty for default based on env/role",
            )

            add_obj(p_PathPattern)

            # outputs
            ListenerRule_Out_String.append("PathPattern=${PathPattern}")
            ListenerRule_Out_Map.update(
                {"PathPattern": get_endvalue(f"{mapname}PathPattern", issub=True)}
            )

        # resources
        for lb in cfg.LoadBalancer:
            LB_ListenerRulesExternalInternal(
                index=str(n), key=v, mapname=mapname, scheme=lb
            )

        # outputs
        o_ListenerRule = Output(mapname)
        o_ListenerRule.Value = Sub(
            ",".join(ListenerRule_Out_String), **ListenerRule_Out_Map
        )

        add_obj(o_ListenerRule)


def LB_ListenersV2ApplicationEC2():
    LB_ListenersEC2()
    for n in ["External", "Internal"]:
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_Http = elbv2.Listener(f"ListenerHttp{n}")
        auto_get_props(r_Http, mapname=f"ListenerV2EC2Http{n}")
        add_obj(r_Http)

        if n == "External":
            r_Https = elbv2.Listener(f"ListenerHttps{n}")
            auto_get_props(r_Https, mapname=f"ListenerV2EC2Https{n}")
            add_obj(r_Https)

    LB_TargetGroupsEC2Application()


def LB_ListenersV2NetworkEC2():
    LB_ListenersEC2()
    for lb in ["External", "Internal"]:
        # resources
        if lb not in cfg.LoadBalancer:
            continue
        for n, v in cfg.Listeners.items():
            mapname = f"Listeners{n}"  # Ex ListenersPort5601
            listener = elbv2.Listener(f"{mapname}{lb}")
            auto_get_props(listener, mapname=f"ListenerV2EC2TCP{lb}", remapname=mapname)
            auto_get_props(listener, mapname=mapname)
            add_obj(listener)

            LB_TargetGroupsEC2Network(lb, mapname)


def LB_ListenersV2ECS():
    for n in ["External", "Internal"]:
        # resources
        if n not in cfg.LoadBalancer:
            continue
        if cfg.ListenerLoadBalancerHttpPort not in ["none", 80]:
            # Enable the relative LB SecurityGroupIngress
            cfg.SecurityGroupIngress[f"LoadBalancerApplicationHttp{n}"][
                "IBOX_ENABLED"
            ] = True
            r_Http = elbv2.Listener(f"ListenerHttp{n}")
            auto_get_props(r_Http, mapname=f"ListenerV2ECSHttp{n}")
            add_obj(r_Http)

        if cfg.ListenerLoadBalancerHttpsPort not in ["none", 443] and n == "External":
            # Enable the relative LB SecurityGroupIngress
            cfg.SecurityGroupIngress[f"LoadBalancerApplicationHttps{n}"][
                "IBOX_ENABLED"
            ] = True
            r_Https = elbv2.Listener(f"ListenerHttps{n}")
            auto_get_props(r_Https, mapname=f"ListenerV2ECSHttps{n}")
            add_obj(r_Https)

    # Resources
    LB_TargetGroupsECS()
    LB_ListenerRules()


def LB_TargetGroupsEC2Application():
    for n in cfg.ElasticLoadBalancingV2TargetGroupEC2:
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_TG = elbv2.TargetGroup(f"TargetGroup{n}")
        auto_get_props(r_TG, mapname=f"ElasticLoadBalancingV2TargetGroupEC2{n}")
        add_obj(r_TG)

        cfg.Alarm[f"TargetEC2{n}5XX"]["IBOX_ENABLED"] = True


def LB_TargetGroupsEC2Network(lb, mapname_listener):
    # resources
    r_TG = elbv2.TargetGroup(f"TargetGroup{mapname_listener}{lb}")
    auto_get_props(
        r_TG,
        mapname=f"ElasticLoadBalancingV2TargetGroupEC2Network",
        remapname=mapname_listener,
    )
    auto_get_props(r_TG, mapname=mapname_listener)
    add_obj(r_TG)


def LB_TargetGroupsECS():
    for n in cfg.ElasticLoadBalancingV2TargetGroupECS:
        # resources
        if n in ["External", "Internal"] and n not in cfg.LoadBalancer:
            continue
        r_TG = elbv2.TargetGroup(f"TargetGroup{n}")
        auto_get_props(r_TG, mapname=f"ElasticLoadBalancingV2TargetGroupECS{n}")
        add_obj(r_TG)

        try:
            cfg.Alarm[f"Target{n}5XX"]["IBOX_ENABLED"] = True
        except Exception:
            pass


def LB_TargetGroupsALB():
    for n in cfg.ElasticLoadBalancingV2TargetGroupALB:
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_TG = elbv2.TargetGroup(f"TargetGroupServiceUnavailable{n}")
        auto_get_props(
            r_TG, mapname=f"ElasticLoadBalancingV2TargetGroupALB{n}", indexname=n
        )

        add_obj(r_TG)


def LB_ElasticLoadBalancingClassicEC2():
    Listeners = LB_ListenersEC2()
    for n, v in cfg.ElasticLoadBalancingLoadBalancer.items():
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_LB = elb.LoadBalancer(f"LoadBalancerClassic{n}")
        auto_get_props(r_LB, mapname=f"ElasticLoadBalancingLoadBalancer{n}")
        r_LB.Listeners = Listeners

        add_obj(r_LB)
        cfg.Alarm[f"Backend{n}5XX"]["IBOX_ENABLED"] = True


def LB_ElasticLoadBalancingApplicationEC2():
    for n, v in cfg.ElasticLoadBalancingV2LoadBalancerAPP.items():
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_LB = elbv2.LoadBalancer(f"LoadBalancerApplication{n}")
        auto_get_props(r_LB, mapname=f"ElasticLoadBalancingV2LoadBalancerAPP{n}")
        add_obj(r_LB)

    LB_ListenersV2ApplicationEC2()


def LB_ElasticLoadBalancingNetworkEC2():
    for n, v in cfg.ElasticLoadBalancingV2LoadBalancerNET.items():
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_LB = elbv2.LoadBalancer(f"LoadBalancerNetwork{n}")
        auto_get_props(r_LB, mapname=f"ElasticLoadBalancingV2LoadBalancerNET{n}")
        add_obj(r_LB)

    LB_ListenersV2NetworkEC2()


def LB_ElasticLoadBalancingEC2(key):
    if not cfg.LoadBalancer:
        return
    enable_recordset("EC2")

    if cfg.LoadBalancerType == "Classic":
        LB_ElasticLoadBalancingClassicEC2()

    if cfg.LoadBalancerType == "Application":
        LB_ElasticLoadBalancingApplicationEC2()

    if cfg.LoadBalancerType == "Network":
        LB_ElasticLoadBalancingNetworkEC2()


def LB_ElasticLoadBalancingALB(key):
    for n, v in cfg.ElasticLoadBalancingV2LoadBalancerALB.items():
        # resources
        if n not in cfg.LoadBalancer:
            continue
        r_LB = elbv2.LoadBalancer(f"LoadBalancerApplication{n}")
        auto_get_props(r_LB, mapname=f"ElasticLoadBalancingV2LoadBalancerALB{n}")

        r_ListenerHttp = elbv2.Listener(f"ListenerHttpDefault{n}")
        auto_get_props(r_ListenerHttp, mapname=f"ListenerV2ALBHttp{n}")

        if n == "External":
            # Https enabled only for External ELB
            r_ListenerHttps = elbv2.Listener(f"ListenerHttpsDefault{n}")
            auto_get_props(r_ListenerHttps, mapname=f"ListenerV2ALBHttps{n}")
            add_obj(r_ListenerHttps)

        add_obj([r_LB, r_ListenerHttp])

    # Resources
    # Create TargetGroups pointing to LambdaServiceUnavailable
    try:
        cfg.ServiceUnavailable
    except Exception:
        pass
    else:
        LB_TargetGroupsALB()


def LB_ElasticLoadBalancingECS(key):
    if not cfg.LoadBalancer:
        return
    # Resources
    LB_ListenersV2ECS()
    enable_recordset("ECS")



import troposphere.ecr as ecr

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


class ECRRepositories(ecr.Repository):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        self.RepositoryName = get_endvalue(self.title)
        self.RepositoryPolicyText = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": [
                        "ecr:GetDownloadUrlForLayer",
                        "ecr:BatchGetImage",
                        "ecr:BatchCheckLayerAvailability",
                        "ecr:PutImage",
                        "ecr:InitiateLayerUpload",
                        "ecr:UploadLayerPart",
                        "ecr:CompleteLayerUpload",
                        "ecr:ListImages",
                        "ecr:DescribeRepositories",
                        "ecr:DescribeImages",
                    ],
                    "Effect": "Allow",
                    "Principal": {"AWS": [Sub("arn:aws:iam::${AWS::AccountId}:root")]},
                    "Sid": "AllowPushPull",
                },
            ],
        }


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


class ECRRepositoryLifecyclePolicy(ecr.LifecyclePolicy):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        LifecyclePolicyText = {
            "rules": [
                {
                    "action": {"type": "expire"},
                    "rulePriority": 1,
                    "selection": {
                        "countNumber": 9500,
                        "countType": "imageCountMoreThan",
                        "tagStatus": "any",
                    },
                    "description": "Images are sorted on pushed_at_time "
                    "(desc), images greater than specified count are expired.",
                }
            ]
        }
        self.LifecyclePolicyText = json.dumps(LifecyclePolicyText)
        self.RegistryId = Ref("AWS::AccountId")


def ECRRepositoryPolicyStatementAccountPull(name):
    policy = {
        "Action": [
            "ecr:GetDownloadUrlForLayer",
            "ecr:BatchGetImage",
            "ecr:BatchCheckLayerAvailability",
            "ecr:ListImages",
            "ecr:DescribeRepositories",
            "ecr:DescribeImages",
        ],
        "Effect": "Allow",
        "Principal": {"AWS": [get_subvalue("arn:aws:iam::${1M}:root", name)]},
        "Sid": "AllowPull",
    }

    return policy


def ECRRepositoryPolicyStatementAccountPush(name):
    policy = {
        "Action": [
            "ecr:PutImage",
            "ecr:InitiateLayerUpload",
            "ecr:UploadLayerPart",
            "ecr:CompleteLayerUpload",
            "ecr:BatchCheckLayerAvailability",
        ],
        "Effect": "Allow",
        "Principal": {"AWS": [get_subvalue("arn:aws:iam::${1M}:root", name)]},
        "Sid": "AllowPush",
    }

    return policy


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


def ECR_Repositories(key):
    PolicyStatementAccounts = []
    for n, v in cfg.EcrAccount.items():
        mapname = f"EcrAccount{n}Id"  # Ex. EcrAccountPrdId
        # conditions
        add_obj(get_condition(mapname, "not_equals", "none"))

        if "Pull" in v["Policy"]:
            PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPull(
                name=mapname
            )
            PolicyStatementAccounts.append(
                If(mapname, PolicyStatementAccount, Ref("AWS::NoValue"))
            )

        if "Push" in v["Policy"]:
            PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPush(
                name=mapname
            )
            PolicyStatementAccounts.append(
                If(mapname, PolicyStatementAccount, Ref("AWS::NoValue"))
            )

    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"
        Repo = ECRRepositories(
            resname, LifecyclePolicy=ECRRepositoryLifecyclePolicy("")
        )
        Repo.RepositoryPolicyText["Statement"].extend(PolicyStatementAccounts)

        add_obj(Repo)



import troposphere.cloudformation as cfm

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    camel_to_snake,
)


def cfn_ecs_cluster():
    init_args = {}

    init_args.update(
        {
            "REPOSITORIES": cfm.InitConfig(
                commands=If(
                    "GPUInstance",
                    {
                        "10-epel-repo": {
                            "command": "amazon-linux-extras install -y epel",
                            "test": "! test -e /etc/yum.repos.d/epel.repo",
                        },
                        # 'cuda-repo': {
                        #    'command': 'yum install -y '
                        #    'http://developer.download.nvidia.com'
                        #    '/compute/cuda/repos/rhel7/x86_64/'
                        #    'cuda-repo-rhel7-10.0.130-1.x86_64.rpm',
                        #    'test': 'test ! -e /etc/yum.repos.d/cuda.repo'
                        # },
                        "nvidia-docker": {
                            "command": "curl -s -L "
                            "https://nvidia.github.io"
                            "/nvidia-docker/amzn2/nvidia-docker.repo | "
                            "tee /etc/yum.repos.d/nvidia-docker.repo",
                            "test": "test ! -e /etc/yum.repos.d/nvidia-docker.repo",
                        },
                    },
                    Ref("AWS::NoValue"),
                ),
                files=If(
                    "GPUInstance",
                    {
                        "/etc/yum.repos.d/epel-nvidia.repo": {
                            "content": Join(
                                "\n",
                                [
                                    "[epel-nvidia]",
                                    "name=negativo17 - Nvidia",
                                    "baseurl=https://negativo17.org"
                                    "/repos/nvidia/epel-7/$basearch/",
                                    "enabled=1",
                                    "skip_if_unavailable=1",
                                    "gpgcheck=1",
                                    "gpgkey=https://negativo17.org"
                                    "/repos/RPM-GPG-KEY-slaanesh",
                                    "enabled_metadata=1",
                                    "metadata_expire=6h",
                                    "type=rpm-md",
                                    "repo_gpgcheck=0",
                                ],
                            )
                        }
                    },
                    Ref("AWS::NoValue"),
                ),
            ),
            "PACKAGES": cfm.InitConfig(
                packages={
                    "yum": {
                        "nfs-utils": [],
                        "make": If("GPUInstance", [], Ref("AWS::NoValue")),
                        "automake": If("GPUInstance", [], Ref("AWS::NoValue")),
                        "gcc": If("GPUInstance", [], Ref("AWS::NoValue")),
                        "gcc-c++": If("GPUInstance", [], Ref("AWS::NoValue")),
                        "kernel-devel": If("GPUInstance", [], Ref("AWS::NoValue")),
                        # 'cuda': If('GPUInstance', [], Ref('AWS::NoValue')),
                        "dkms-nvidia": If("GPUInstance", [], Ref("AWS::NoValue")),
                        "nvidia-driver-cuda": If(
                            "GPUInstance", [], Ref("AWS::NoValue")
                        ),
                        "nvidia-docker2": If("GPUInstance", [], Ref("AWS::NoValue")),
                    }
                }
            ),
            "SERVICES": cfm.InitConfig(
                files={
                    "/etc/ecs/ecs.config": {
                        "content": Join(
                            "\n",
                            [Sub("ECS_CLUSTER=${Cluster}")]
                            + [
                                get_subvalue(
                                    "%s=${1M}" % camel_to_snake(n),
                                    f"ECSClusterBaseAgentCfg{n}",
                                )
                                for n in cfg.ECSClusterBaseAgentCfg
                            ]
                            + [
                                If(
                                    "GPUInstance",
                                    "ECS_DISABLE_PRIVILEGED=false",
                                    Ref("AWS::NoValue"),
                                ),
                            ],
                        )
                    },
                    "/etc/docker/daemon.json": If(
                        "GPUInstance",
                        {
                            "content": Join(
                                "\n",
                                [
                                    "{",
                                    '  "default-runtime": "nvidia",',
                                    '  "runtimes": {',
                                    '    "nvidia": {',
                                    '      "path": "/usr/bin/nvidia-container-runtime",',
                                    '      "runtimeArgs": []',
                                    "    }",
                                    "  }",
                                    "}",
                                ],
                            )
                        },
                        Ref("AWS::NoValue"),
                    ),
                },
                commands={
                    "01-kern-modules": If(
                        "GPUInstance",
                        {
                            "command": "rmmod nvidia_modeset;rmmod nvidia_uvm;"
                            "rmmod nvidia;modprobe nvidia;modprobe nvidia_uvm;true"
                        },
                        Ref("AWS::NoValue"),
                    ),
                    "02-restart-docker": If(
                        "GPUInstance",
                        {"command": "pkill -SIGHUP dockerd"},
                        Ref("AWS::NoValue"),
                    ),
                },
            ),
        }
    )

    # aws ecs ami by default run postfix, use yaml cfg Postfix: false to
    # stop it
    try:
        cfg.Postfix
    except Exception:
        pass
    else:
        # Parameter
        P_Postfix = Parameter("Postfix")
        P_Postfix.Description = "Disable Postfix Service"
        # Condition
        C_Postfix = get_condition("Postfix", "equals", "no")
        # Output
        O_Postfix = Output("Postfix")
        O_Postfix.Value = get_endvalue("Postfix")

        add_obj(
            [
                P_Postfix,
                C_Postfix,
                O_Postfix,
            ]
        )

        init_args["SERVICES"].services = {
            "sysvinit": {
                "postfix": If(
                    "Postfix",
                    {"enabled": "false", "ensureRunning": "false"},
                    Ref("AWS::NoValue"),
                )
            }
        }

    return init_args



import troposphere.autoscaling as asg
import troposphere.ec2 as ec2
import troposphere.cloudformation as cfm
import troposphere.policies as pol
import troposphere.applicationautoscaling as aas

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    import_user_data,
)
from .cfn import *

try:
    from cfnExt import *
except ModuleNotFoundError:
    pass


class ASLaunchTemplateData(ec2.LaunchTemplateData):
    def __init__(self, title, UserDataApp, **kwargs):
        super().__init__(title, **kwargs)
        auto_get_props(self, "LaunchTemplateData")
        user_data = [
            "#!/bin/bash",
            "PATH=/opt/aws/bin:/usr/local/bin:$PATH",
            "export BASH_ENV=/etc/profile.d/ibox_env.sh",
            "export ENV=$BASH_ENV",
            "yum -C list installed aws-cfn-bootstrap || yum install -y aws-cfn-bootstrap",
            "if ( ! which cfn-init );then",
            "  yum install -y python3-pip chkconfig",
            "  pip3 install https://s3.amazonaws.com/cloudformation-examples/aws-cfn-bootstrap-py3-latest.tar.gz",
            "  cp /usr/local/init/systemd/cfn-hup.service /etc/systemd/system/ && systemctl daemon-reload",
            "fi",
            Sub("".join(UserDataApp)),
            Sub(
                "cfn-init -v --stack ${AWS::StackName} --role ${RoleInstance}"
                " --resource LaunchTemplate --region ${AWS::Region}"
            ),
            If(
                "DoNotSignal",
                Ref("AWS::NoValue"),
                Sub(
                    "cfn-signal -e $? --stack ${AWS::StackName}"
                    " --role ${RoleInstance} --resource AutoScalingGroup"
                    " --region ${AWS::Region}"
                ),
            ),
            "rm /var/lib/cloud/instance/sem/config_scripts_user\n",
        ]

        user_data = Join("\n", user_data)

        try:
            # look for external file
            user_data = import_user_data(
                getattr(cfg, "IBOX_ROLE_EX", getattr(cfg, "envrole"))
            )
        except Exception:
            pass
        else:
            if "cfn-init" not in user_data:
                cfg.use_cfn_init = False
            user_data = Join("", user_data)

        try:
            # look for bottlerocket key
            cfg.BottleRocket
        except Exception:
            # use standard cfg
            self.UserData = Base64(user_data)
        else:
            # use if condition with both bottlerocket custom and standard cfg
            self.UserData = Base64(
                If(
                    "BottleRocket",
                    Join(
                        "\n",
                        [
                            get_endvalue(f"BottleRocketUserData{n}Line")
                            for n in cfg.BottleRocketUserData
                        ],
                    ),
                    user_data,
                ),
            )


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


class ASInitConfigSets(cfm.InitConfigSets):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if cfg.Apps:
            CODEDEPLOY = If("DeploymentGroup", "CODEDEPLOY", Ref("AWS::NoValue"))
        else:
            CODEDEPLOY = Ref("AWS::NoValue")

        CWAGENT = If("CloudWatchAgent", "CWAGENT", Ref("AWS::NoValue"))

        if getattr(cfg, "IBOX_LAUNCH_TEMPLATE_NO_WAIT_ELB_HEALTH", False):
            ELBWAITER = Ref("AWS::NoValue")
        elif cfg.LoadBalancer or cfg.LoadBalancer:
            ELBWAITER = "ELBWAITER"
        else:
            ELBWAITER = Ref("AWS::NoValue")

        self.data = {
            "default": [
                "REPOSITORIES",
                "PACKAGES",
                "SETUP",
                CODEDEPLOY,
                "SERVICES",
                CWAGENT,
                ELBWAITER,
            ],
            "buildami": [
                "REPOSITORIES",
                "PACKAGES",
            ],
            "buildamifull": [
                "REPOSITORIES",
                "PACKAGES",
                "SETUP",
            ],
        }


class ASInitConfigSetup(cfm.InitConfig):
    def __init__(self, **kwargs):
        self.ibox_env_app = ""
        super(ASInitConfigSetup, self).__init__(**kwargs)

    def setup(self):
        self.files = {
            "/etc/profile.d/ibox_env.sh": {
                "content": Join(
                    "",
                    [
                        "#setup ibox environment variables\n",
                        "export Env=",
                        Ref("Env"),
                        "\n",
                        "export EnvAbbr=",
                        Ref("EnvShort"),
                        "\n",
                        "export EnvRegion=",
                        Ref("AWS::Region"),
                        "\n",
                        "export EnvAccountId=",
                        Ref("AWS::AccountId"),
                        "\n",
                        "export EnvRole=",
                        Ref("EnvRole"),
                        "\n",
                        "export EnvBrand=",
                        cfg.BrandDomain,
                        "\n",
                        "export EnvStackName=",
                        Ref("AWS::StackName"),
                        "\n",
                    ]
                    + self.ibox_env_app,
                )
            },
            "/etc/cfn/cfn-hup.conf": {
                "content": Join(
                    "",
                    [
                        "[main]\n",
                        "stack=",
                        Ref("AWS::StackId"),
                        "\n",
                        "region=",
                        Ref("AWS::Region"),
                        "\n",
                        "role=",
                        Ref("RoleInstance"),
                        "\n",
                        "interval=5\n",
                    ],
                ),
                "mode": "000400",
                "owner": "root",
                "group": "root",
            },
            "/etc/cfn/hooks.d/cfn-auto-reloader.conf": {
                "content": Join(
                    "",
                    [
                        "[cfn-auto-reloader-hook]\n",
                        "triggers=post.add, post.update\n",
                        "path=Resources.LaunchTemplate"
                        ".Metadata.CloudFormationInitVersion\n",
                        "action=/opt/aws/bin/cfn-init -v",
                        " --stack ",
                        Ref("AWS::StackName"),
                        " --role ",
                        Ref("RoleInstance"),
                        " --resource LaunchTemplate",
                        " --region ",
                        Ref("AWS::Region"),
                        "\n",
                        "runas=root\n",
                    ],
                )
            },
            # '/usr/local/bin/chamber': {
            #     'mode': '000755',
            #     'source': Sub(
            #         f'https://{cfg.BucketNameAppRepository}'
            #         '.s3.${AWS::Region}.amazonaws.com/ibox-tools/chamber'),
            #     'owner': 'root',
            #     'group': 'root',
            # },
        }
        self.commands = {
            "01_disk": {
                "command": Join(
                    "",
                    [
                        "n=0\n",
                        "for disk in /dev/xvd[b-d]; do\n",
                        '  [ -b "$disk" ] || continue\n',
                        '  file -s "$disk" | grep -q "ext[34] filesystem" || ',
                        "  { mkfs.ext4 $disk || continue; }\n",
                        "  mkdir -p /media/ephemeral${n} && ",
                        "  mount $disk /media/ephemeral${n}\n",
                        "  n=$(($n+1))\n",
                        "done",
                    ],
                )
            },
            "02_disk_additional": If(
                "LaunchTemplateDataBlockDeviceMappingsAdditionalStorageMount",
                {
                    "command": get_subvalue(
                        'file -s ${1M}1 | grep -q "ext[34] filesystem" ||'
                        " { parted -s ${1M} mklabel gpt &&"
                        " parted -s ${1M} mkpart primary ext2 1 ${2M}G &&"
                        " mkfs.ext4 ${1M}1 || continue; }\nmkdir -p /data &&"
                        " mount ${1M}1 /data",
                        [
                            (
                                "LaunchTemplateDataBlockDeviceMappings"
                                "AdditionalStorageDeviceName"
                            ),
                            (
                                "LaunchTemplateDataBlockDeviceMappings"
                                "AdditionalStorageEbsVolumeSize"
                            ),
                        ],
                    )
                },
                Ref("AWS::NoValue"),
            ),
            "03_efs_mounts": If(
                "EfsMounts",
                {
                    "command": Join(
                        "",
                        [
                            "for mounts in ",
                            Join(" ", Ref("EfsMounts")),
                            ";do\n",
                            '  mkdir -p "/media/${mounts}"\n',
                            '  mountpoint -q "/media/${mounts}" ||',
                            "    mount -t nfs4 -o nfsvers=4,rsize=1048576,"
                            "wsize=1048576,hard,timeo=600,retrans=2 ",
                            "    efs-${mounts}.",
                            cfg.HostedZoneNamePrivate,
                            ":/ ",
                            "    /media/${mounts}\n",
                            "done",
                        ],
                    ),
                },
                Ref("AWS::NoValue"),
            ),
            "04_rmdir_tmp_ibox": {"command": "rm -fr /tmp/ibox"},
        }
        self.services = {
            "sysvinit": {
                "cfn-hup": {
                    "enabled": "false",
                    "ensureRunning": "true",
                    "files": [
                        "/etc/cfn/cfn-hup.conf",
                        "/etc/cfn/hooks.d/cfn-auto-reloader.conf",
                    ],
                }
            }
        }


class ASInitConfigCodeDeploy(cfm.InitConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.files = {
            "/etc/codedeploy-agent/conf/codedeployagent.yml": {
                "content": Join(
                    "",
                    [
                        "---\n",
                        ":log_aws_wire: false\n",
                        ":log_dir: '/var/log/aws/codedeploy-agent/'\n",
                        ":pid_dir: '/opt/codedeploy-agent/state/.pid/'\n",
                        ":program_name: codedeploy-agent\n",
                        ":root_dir: '/opt/codedeploy-agent/deployment-root'\n",
                        ":verbose: false\n",
                        ":wait_between_runs: 1\n",
                        ":proxy_uri:\n",
                        ":max_revisions: 2\n",
                    ],
                )
            },
            "/tmp/codedeployinstall.sh": {
                "source": Sub(
                    "https://aws-codedeploy-${AWS::Region}"
                    ".s3.amazonaws.com/latest/install"
                ),
                "mode": "000700",
            },
        }
        self.commands = {
            "03_codedeploy-install-and-run": {
                "command": "/tmp/codedeployinstall.sh auto"
            }
        }
        self.services = {
            "sysvinit": {
                "codedeploy-agent": {
                    "enabled": "false",
                    "ensureRunning": "true",
                    "files": ["/etc/codedeploy-agent/conf/codedeployagent.yml"],
                }
            }
        }


class ASInitConfigCloudWatchAgent(cfm.InitConfig):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        self.packages = {
            "rpm": {
                "amazon-cloudwatch-agent": (
                    "https://s3.amazonaws.com"
                    "/amazoncloudwatch-agent/amazon_linux/amd64/latest/"
                    "amazon-cloudwatch-agent.rpm"
                )
            }
        }
        self.files = {
            "/opt/aws/amazon-cloudwatch-agent/etc/"
            "amazon-cloudwatch-agent.json": {
                "content": Join(
                    "",
                    [
                        "{\n",
                        '  "metrics": {\n',
                        '    "append_dimensions": {\n',
                        '      "AutoScalingGroupName": '
                        '"${aws:AutoScalingGroupName}",\n',
                        '      "InstanceId": "${!aws:InstanceId}"\n',
                        "    },\n",
                        '    "aggregation_dimensions": [\n',
                        '      ["AutoScalingGroupName"]\n',
                        "    ],\n",
                        '    "metrics_collected": {\n',
                        '      "mem": {\n',
                        '        "measurement": [\n',
                        '          "mem_used_percent"\n',
                        "        ]\n",
                        "      },\n",
                        '      "disk": {\n',
                        '        "resources": [\n',
                        '          "/"\n',
                        "        ],\n",
                        '        "measurement": [\n',
                        "          {\n",
                        '            "name": "disk_used_percent",\n',
                        '            "rename": "root_disk_used_percent"\n',
                        "          }\n",
                        "        ],\n",
                        '        "drop_device": true\n',
                        "      }\n",
                        "    }\n",
                        "  }\n",
                        "}\n",
                    ],
                ),
            },
        }
        self.services = {
            "sysvinit": {
                "amazon-cloudwatch-agent": {
                    "ensureRunning": "true",
                    "files": [
                        "/opt/aws/amazon-cloudwatch-agent/etc/"
                        "amazon-cloudwatch-agent.json"
                    ],
                }
            }
        }


class ASInitConfigApps(cfm.InitConfig):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title  # Ex. Apps1
        reponame = f"{name}RepoName"
        n = name.replace("Apps", "")
        envappversion = f"EnvApp{n}Version"

        self.sources = {
            "/tmp/ibox/": If(
                "DeployRevision",
                Ref("AWS::NoValue"),
                get_subvalue(
                    "https://%s.s3-${AWS::Region}.amazonaws.com/"
                    "${1M}/${1M}-${%s}.tar.gz"
                    % (cfg.BucketNameAppRepository, envappversion),
                    reponame,
                    "",
                ),
            )
        }
        self.commands = {
            "01_setup": If(
                "DeployRevision",
                Ref("AWS::NoValue"),
                {
                    "command": get_subvalue(
                        "EnvAppVersion=${%s} EnvRepoName=${1M} "
                        "/tmp/ibox/bin/setup.sh" % envappversion,
                        reponame,
                    )
                },
            ),
            "02_setup_reboot_codedeploy": If(
                "DeployRevision",
                {
                    "command": get_subvalue(
                        "EnvAppVersion=${%s} EnvRepoName=${1M} "
                        "/opt/ibox/${1M}/live/bin/setup.sh" % envappversion,
                        reponame,
                    ),
                    "test": "test -e /var/lib/cloud/instance/sem/"
                    "config_ssh_authkey_fingerprints",
                },
                Ref("AWS::NoValue"),
            ),
            "03_rmdir_tmp_ibox": {"command": "rm -fr /tmp/ibox"},
        }


class ASInitConfigAppsBuildAmi(ASInitConfigApps):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        for n, v in self.sources.items():
            if not isinstance(v, dict):
                if "Fn::If" in v.data:
                    self.sources[n] = v.data["Fn::If"][2]

        for n, v in self.commands.items():
            if not isinstance(v, dict):
                if "Fn::If" in v.data:
                    self.commands[n] = v.data["Fn::If"][2]


class ASInitConfigELBClassic(cfm.InitConfig):
    def __init__(self, scheme, **kwargs):
        super().__init__(**kwargs)

        self.commands = {
            "ELBClassicInternalHealthCheck": {
                "command": Join(
                    "",
                    [
                        'until [ "$state" = \'"InService"\' ]; do',
                        "  state=$(aws --region ",
                        Ref("AWS::Region"),
                        " elb describe-instance-health",
                        "  --load-balancer-name ",
                        Ref(f"LoadBalancerClassic{scheme}"),
                        "  --instances $(curl -s "
                        "http://169.254.169.254/latest/meta-data/instance-id)",
                        "  --query InstanceStates[0].State);",
                        "  sleep 10;",
                        "done",
                    ],
                )
            }
        }


class ASInitConfigELBApplication(cfm.InitConfig):
    def __init__(self, scheme, **kwargs):
        super().__init__(**kwargs)

        self.commands = {
            "ELBApplicationExternalHealthCheck": {
                "command": Join(
                    "",
                    [
                        'until [ "$state" = \'"healthy"\' ]; do',
                        "  state=$(aws --region ",
                        Ref("AWS::Region"),
                        " elbv2 describe-target-health",
                        "  --target-group-arn ",
                        Ref(f"TargetGroup{scheme}"),
                        "  --targets Id=$(curl -s "
                        "http://169.254.169.254/latest/meta-data/instance-id)",
                        "  --query " "TargetHealthDescriptions[0].TargetHealth.State);",
                        "  sleep 10;",
                        "done",
                    ],
                )
            }
        }


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


def AS_ScalingPolicies(key):
    Out_String = []
    Out_Map = {}
    for n, v in getattr(cfg, key).items():
        if not v["IBOX_ENABLED"]:
            continue

        resname = f"{key}{n}"

        # resources
        if key == "AutoScalingScalingPolicy":
            r_Policy = asg.ScalingPolicy(resname)
        else:
            r_Policy = aas.ScalingPolicy(resname)

        auto_get_props(r_Policy)
        add_obj(r_Policy)

        # for tracking create output
        if v["PolicyType"] == "TargetTrackingScaling":
            # Autoscaling
            if "TargetTrackingConfiguration" in v:
                TargetTrackingConfigurationName = "TargetTrackingConfiguration"
            # Application Autoscaling
            elif "TargetTrackingScalingPolicyConfiguration" in v:
                TargetTrackingConfigurationName = (
                    "TargetTrackingScalingPolicyConfiguration"
                )

            basename = f"{resname}{TargetTrackingConfigurationName}"

            # outputs
            if v["Type"] == "Cpu" or (
                v["Type"] == "Custom"
                and v[TargetTrackingConfigurationName]["CustomizedMetricSpecification"][
                    "MetricName"
                ]
                == "CPUUtilization"
            ):
                # Use Cpu Metric
                Out_String.append("Cpu${Statistic}:${Cpu}")

                if v["Type"] == "Custom":
                    statistic = get_endvalue(
                        f"{basename}" "CustomizedMetricSpecificationStatistic"
                    )
                else:
                    statistic = ""

                Out_Map.update(
                    {
                        "Statistic": statistic,
                        "Cpu": get_endvalue(f"{basename}TargetValue"),
                    }
                )

    if Out_String:
        # Outputs
        O_Policy = Output(key, Value=Sub(",".join(Out_String), **Out_Map))

        add_obj(O_Policy)


def AS_LaunchTemplate():
    cfg.use_cfn_init = True
    InitConfigSets = ASInitConfigSets()

    CfmInitArgs = {}
    IBoxEnvApp = []
    Tags_List = []
    UserDataApp = []

    for n in cfg.Apps:
        name = f"Apps{n}"  # Ex. Apps1
        envname = f"EnvApp{n}Version"  # Ex EnvApp1Version
        reponame = f"{name}RepoName"  # Ex Apps1RepoName

        UserDataApp.extend(["#${%s}\n" % envname])

        p_EnvAppVersion = Parameter(
            envname,
            Description=f"Application {n} version",
            AllowedPattern="^[a-zA-Z0-9-_.]*$",
        )

        p_AppsRepoName = Parameter(
            reponame,
            Description=f"App {n} Repo Name - empty for default based on env/role",
            AllowedPattern="^[a-zA-Z0-9-_.]*$",
        )

        # parameters
        add_obj(
            [
                p_EnvAppVersion,
                p_AppsRepoName,
            ]
        )

        # conditions
        add_obj(
            {
                name: And(
                    Not(Equals(Ref(envname), "")),
                    Not(get_condition("", "equals", "None", reponame)),
                )
            }
        )

        InitConfigApps = ASInitConfigApps(name)
        CfmInitArgs[name] = InitConfigApps

        InitConfigAppsBuilAmi = ASInitConfigAppsBuildAmi(name)
        # AUTOSPOT - Let cfn-init always prepare instances on boot
        # CfmInitArgs[name + 'BuildAmi'] = InitConfigAppsBuilAmi
        CfmInitArgs[name] = InitConfigAppsBuilAmi

        IBoxEnvApp.extend(
            [
                f"export EnvApp{n}Version=",
                Ref(envname),
                "\n",
                f"export EnvRepo{n}Name=",
                get_endvalue(reponame),
                "\n",
            ]
        )

        InitConfigSetsApp = If(name, name, Ref("AWS::NoValue"))
        InitConfigSetsAppBuilAmi = If(name, f"{name}BuildAmi", Ref("AWS::NoValue"))
        IndexSERVICES = InitConfigSets.data["default"].index("SERVICES")
        InitConfigSets.data["default"].insert(IndexSERVICES, InitConfigSetsApp)
        # AUTOSPOT - Let cfn-init always prepare instances on boot
        # InitConfigSets.data['buildamifull'].append(
        #    InitConfigSetsAppBuilAmi)
        InitConfigSets.data["buildamifull"].append(InitConfigSetsApp)

        Tags_List.append(asg.Tag(envname, Ref(envname), True))

        # outputs
        Output_app = Output(envname, Value=Ref(envname))
        Output_repo = Output(reponame, Value=get_endvalue(reponame))

        add_obj([Output_app, Output_repo])

    InitConfigSetup = ASInitConfigSetup()
    InitConfigSetup.ibox_env_app = IBoxEnvApp
    InitConfigSetup.setup()

    InitConfigCodeDeploy = ASInitConfigCodeDeploy()

    CfmInitArgs["SETUP"] = InitConfigSetup
    CfmInitArgs["CWAGENT"] = ASInitConfigCloudWatchAgent("")

    if cfg.CodeDeploy:
        CfmInitArgs["CODEDEPLOY"] = InitConfigCodeDeploy

    if not getattr(cfg, "IBOX_LAUNCH_TEMPLATE_NO_WAIT_ELB_HEALTH", False):
        for lb in cfg.LoadBalancer:
            # LoadBalancerClassic
            if cfg.LoadBalancerType == "Classic":
                InitConfigELB = ASInitConfigELBClassic(scheme=lb)
                CfmInitArgs["ELBWAITER"] = InitConfigELB

            # LoadBalancerApplication
            if cfg.LoadBalancerType == "Application":
                InitConfigELB = ASInitConfigELBApplication(scheme=lb)
                CfmInitArgs["ELBWAITER"] = InitConfigELB

            # LoadBalancerNetwork
            if cfg.LoadBalancerType == "Network":
                for k in cfg.Listeners:
                    InitConfigELB = ASInitConfigELBApplication(
                        scheme=f"TargetGroupListeners{k}{lb}"
                    )
                    CfmInitArgs["ELBWAITER"] = InitConfigELB

    if getattr(cfg, "IBOX_LAUNCH_TEMPLATE_NO_SG_EXTRA", False):
        SecurityGroups = []
    else:
        SecurityGroups = cfg.SecurityGroupsImport

    # Resources
    R_LaunchTemplate = ec2.LaunchTemplate(
        "LaunchTemplate",
        LaunchTemplateName=Sub("${AWS::StackName}-${EnvRole}"),
        LaunchTemplateData=ASLaunchTemplateData(
            "LaunchTemplateData", UserDataApp=UserDataApp
        ),
    )
    R_LaunchTemplate.LaunchTemplateData.NetworkInterfaces[0].Groups.extend(
        SecurityGroups
    )

    # Import role specific cfn definition
    try:
        # Do not use role but direct cfg yaml configuration (ecs + cluster)
        cfn_envrole = f"cfn_{cfg.IBOX_ROLE_EX}"
    except Exception:
        cfn_envrole = f"cfn_{cfg.envrole}"
    cfn_envrole = cfn_envrole.replace("-", "_")
    if cfn_envrole in globals():  # Ex cfn_client_portal
        CfnRole = globals()[cfn_envrole]()
        CfmInitArgs.update(CfnRole)

    if cfg.use_cfn_init:
        R_LaunchTemplate.Metadata = cfm.Metadata(
            {
                "CloudFormationInitVersion": If(
                    "CloudFormationInit",
                    Ref("EnvStackVersion"),
                    Ref("AWS::NoValue"),
                )
            },
            cfm.Init(InitConfigSets, **CfmInitArgs),
            cfm.Authentication(
                {
                    "CfnS3Auth": cfm.AuthenticationBlock(
                        type="S3",
                        buckets=[
                            Sub(cfg.BucketNameAppRepository),
                            Sub(cfg.BucketNameAppData),
                        ],
                        roleName=Ref("RoleInstance"),
                    )
                }
            ),
        )

    add_obj(R_LaunchTemplate)

    Tags = asg.Tags()
    Tags.tags = Tags_List
    return Tags


def AS_Autoscaling(key):
    LoadBalancers = []
    TargetGroups = []
    for n in cfg.LoadBalancer:
        if cfg.LoadBalancerType == "Classic":
            LoadBalancers.append(Ref(f"LoadBalancerClassic{n}"))

        if cfg.LoadBalancerType == "Application":
            TargetGroups.append(Ref(f"TargetGroup{n}"))

        if cfg.LoadBalancerType == "Network":
            for k in cfg.Listeners:
                TargetGroups.append(Ref(f"TargetGroupListeners{k}{n}"))

    # Resources
    LaunchTemplateTags = AS_LaunchTemplate()

    R_ASG = asg.AutoScalingGroup(
        "AutoScalingGroupBase",
        LoadBalancerNames=LoadBalancers,
        TargetGroupARNs=TargetGroups,
    )

    auto_get_props(R_ASG)

    R_ASG.Tags += LaunchTemplateTags

    add_obj([R_ASG])



import troposphere.cloudfront as clf

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    change_obj_data,
    clf_compute_order,
)


class CFOriginAccessIdentity(clf.CloudFrontOriginAccessIdentity):
    def __init__(self, title, comment, **kwargs):
        super().__init__(title, **kwargs)
        self.CloudFrontOriginAccessIdentityConfig = (
            clf.CloudFrontOriginAccessIdentityConfig(Comment=comment)
        )


def cache_behavior_process():
    for n, v in cfg.CloudFrontCacheBehaviors.items():
        resname = f"CloudFrontCacheBehaviors{n}"
        # Use CachePolicyId/OriginRequestPolicyId or legacy mode
        if "CachePolicyId" in v:
            for k in ["DefaultTTL", "MaxTTL", "MinTTL", "ForwardedValues"]:
                try:
                    del getattr(cfg, resname)[k]
                except Exception:
                    pass


def origin_process():
    for n, v in cfg.CloudFrontOrigins.items():
        resname = f"CloudFrontOrigins{n}"

        if "OriginAccessIdentity" in v.get("S3OriginConfig", []):
            del getattr(cfg, resname)["CustomOriginConfig"]
        else:
            del getattr(cfg, resname)["S3OriginConfig"]


def CF_CloudFront(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"
        # Resources
        R_CloudFrontDistribution = clf.Distribution(resname)

        # process cache_behavior and origin
        cache_behavior_process()
        origin_process()

        # Automatically compute Behaviour Order based on PathPattern
        cfg.dbg_clf_compute_order = {}
        sortedcachebehaviors = sorted(
            cfg.CloudFrontCacheBehaviors.items(),
            key=lambda x_y: clf_compute_order(x_y[1]["PathPattern"]),
        )
        cfg.CloudFrontCacheBehaviors = {x[0]: x[1] for x in sortedcachebehaviors}

        if cfg.debug:
            print("##########CLF_COMPUTE_ORDER#########START#######")
            for n, v in iter(
                sorted(cfg.dbg_clf_compute_order.items(), key=lambda x_y: x_y[1])
            ):
                print(f"{n} {v}\n")
            print("##########CLF_COMPUTE_ORDER#########END#######")

        auto_get_props(R_CloudFrontDistribution)
        add_obj(R_CloudFrontDistribution)



import troposphere.s3 as s3

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    get_dictvalue,
)
from .iam import IAMPolicyBucketReplica, IAMPolicyStatement
from .cloudfront import CFOriginAccessIdentity


class S3Bucket(s3.Bucket):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title  # Ex. BucketStatic
        # need to specify key cause there is a problem regarind Bucket Names
        # key in common.yml and Bucket dict (RP_to_cfg)
        auto_get_props(self, key=key)
        self.Condition = name
        self.BucketName = Sub(bucket_name)
        self.CorsConfiguration = If(
            f"{name}Cors",
            s3.CorsConfiguration(
                CorsRules=[
                    s3.CorsRules(
                        AllowedHeaders=["Authorization"],
                        AllowedMethods=["GET"],
                        AllowedOrigins=["*"],
                        MaxAge=3000,
                    )
                ]
            ),
            Ref("AWS::NoValue"),
        )

        self.VersioningConfiguration = If(
            f"{name}Versioning",
            s3.VersioningConfiguration(Status=get_endvalue(f"{name}Versioning")),
            Ref("AWS::NoValue"),
        )


class S3BucketPolicy(s3.BucketPolicy):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)

        if "Condition" in key:
            self.Condition = key["Condition"]
        self.PolicyDocument = {
            "Version": "2012-10-17",
        }


def S3BucketPolicyStatementBase(bucket):
    statements = []
    statements.append(
        {
            "Action": ["s3:GetBucketLocation"],
            "Effect": "Allow",
            "Resource": Sub("arn:aws:s3:::%s" % bucket_name),
            "Principal": {"AWS": Sub("arn:aws:iam::${AWS::AccountId}:root")},
            "Sid": "Base",
        }
    )

    return statements


def S3BucketPolicyStatementReplica(bucket, resource):
    statements = []
    if_statements = []
    condition = f"{bucket}PolicyStatementReplicaPrincipal"
    statements.append(
        {
            "Action": [
                "s3:ReplicateObject",
                "s3:ReplicateDelete",
                "s3:ObjectOwnerOverrideToBucketOwner",
            ],
            "Effect": "Allow",
            "Resource": resource,
            "Principal": {
                "AWS": [
                    get_subvalue(
                        "arn:aws:iam::${1M}:root",
                        f"{bucket}PolicyStatementReplicaPrincipal",
                    )
                ]
            },
            "Sid": "AllowReplica",
        }
    )

    for s in statements:
        if_statements.append(If(condition, s, Ref("AWS::NoValue")))

    return if_statements


def S3BucketPolicyStatementAllowGetObject(bucket, principal, sid):
    statements = []
    statements.append(
        {
            "Action": ["s3:GetObject"],
            "Effect": "Allow",
            "Resource": [Sub("arn:aws:s3:::%s/*" % bucket_name)],
            "Principal": {"AWS": principal},
            "Sid": sid,
        },
    )

    return statements


def S3BucketPolicyStatementSes(bucket):
    statements = []
    statements.append(
        {
            "Action": ["s3:PutObject"],
            "Effect": "Allow",
            "Resource": [Sub("arn:aws:s3:::%s/*" % bucket_name)],
            "Principal": {"Service": "ses.amazonaws.com"},
            "Condition": {
                "StringEquals": {"aws:Referer": Ref("AWS::AccountId")},
            },
            "Sid": "AllowSES",
        },
    )

    return statements


def S3BucketPolicyStatementRead(bucket, principal):
    statements = []
    if_statements = []
    condition = f"{bucket}PolicyRead"
    statements.append(
        {
            "Action": [
                "s3:ListBucket",
                "s3:GetBucketLocation",
                "s3:ListBucketMultipartUploads",
                "s3:ListBucketVersions",
            ],
            "Effect": "Allow",
            "Resource": [Sub("arn:aws:s3:::%s" % bucket_name)],
            "Principal": {"AWS": principal},
            "Sid": "AllowListBucket",
        }
    )
    statements.append(
        {
            "Action": ["s3:GetObject", "s3:ListMultipartUploadParts"],
            "Effect": "Allow",
            "Resource": [Sub("arn:aws:s3:::%s/*" % bucket_name)],
            "Principal": {"AWS": principal},
            "Sid": "AllowGetObject",
        }
    )

    for s in statements:
        if_statements.append(If(condition, s, Ref("AWS::NoValue")))

    return if_statements


def S3BucketPolicyStatementWrite(bucket, principal):
    statements = []
    if_statements = []
    condition = f"{bucket}PolicyWrite"
    statements.append(
        {
            "Action": [
                "s3:Put*",
            ],
            "Effect": "Allow",
            "Resource": [Sub("arn:aws:s3:::%s/*" % bucket_name)],
            "Principal": {"AWS": principal},
            "Sid": "AllowPut",
        }
    )

    for s in statements:
        if_statements.append(If(condition, s, Ref("AWS::NoValue")))

    return if_statements


def S3_Buckets(key):
    global bucket_name

    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"
        name = n
        bucket_name = getattr(cfg, f"{key}Name{n}")

        PolicyReadConditions = []
        PolicyReadPrincipal = []

        for m, w in v["AccountsRead"].items():
            accountread_name = f"{resname}AccountsRead{m}"
            # conditions
            add_obj(get_condition(accountread_name, "not_equals", "none"))

            PolicyReadConditions.append(Condition(accountread_name))
            PolicyReadPrincipal.append(
                If(
                    accountread_name,
                    get_subvalue("arn:aws:iam::${1M}:root", accountread_name),
                    Ref("AWS::NoValue"),
                )
            )

        # conditions
        if PolicyReadConditions:
            c_PolicyRead = {
                f"{resname}PolicyRead": Or(
                    Equals("1", "0"), Equals("1", "0"), *PolicyReadConditions
                )
            }
        else:
            c_PolicyRead = {f"{resname}PolicyRead": Equals("True", "False")}

        PolicyWriteConditions = []
        PolicyWritePrincipal = []

        for m, w in v["AccountsWrite"].items():
            accountwrite_name = f"{resname}AccountsWrite{m}"
            # conditions
            add_obj(get_condition(accountwrite_name, "not_equals", "none"))

            PolicyWriteConditions.append(Condition(accountwrite_name))
            PolicyWritePrincipal.append(
                If(
                    accountwrite_name,
                    get_subvalue("arn:aws:iam::${1M}:root", accountwrite_name),
                    Ref("AWS::NoValue"),
                )
            )

        # conditions
        if PolicyWriteConditions:
            c_PolicyWrite = {
                f"{resname}PolicyWrite": Or(
                    Equals("1", "0"), Equals("1", "0"), *PolicyWriteConditions
                )
            }
        else:
            c_PolicyWrite = {f"{resname}PolicyWrite": Equals("True", "False")}

        c_Create = get_condition(resname, "equals", "yes", f"{resname}Create")

        c_Versioning = get_condition(f"{resname}Versioning", "equals", "Enabled")

        c_Cors = get_condition(f"{resname}Cors", "equals", "yes")

        c_Replica = {
            f"{resname}Replica": And(
                Condition(resname),
                get_condition("", "equals", "yes", f"{resname}ReplicationEnabled"),
            )
        }
        c_ReplicaPolicyStatement = get_condition(
            f"{resname}PolicyStatementReplicaPrincipal", "not_equals", "none"
        )

        add_obj(
            [
                c_PolicyRead,
                c_PolicyWrite,
                c_Create,
                c_Versioning,
                c_Cors,
                c_Replica,
                c_ReplicaPolicyStatement,
            ]
        )

        # resources
        r_Bucket = S3Bucket(resname, key=v)

        Replica_Rules = []
        for m, w in v["Replication"]["ConfigurationRules"].items():
            replica_name = f"{resname}ReplicationConfigurationRules{m}"

            # parameters
            p_replicabucket = Parameter(
                f"{replica_name}DestinationBucket",
                Description="Replica Destination Bucket - empty for default based on Env/Roles/Region",
            )
            add_obj(p_replicabucket)

            # conditions
            add_obj(
                [
                    get_condition(
                        f"{replica_name}DestinationBucket", "not_equals", "none"
                    )
                ]
            )

            # resources
            rule = s3.ReplicationConfigurationRules(replica_name, Status="Enabled")
            auto_get_props(rule)

            Replica_Rules.append(
                If(f"{replica_name}DestinationBucket", rule, Ref("AWS::NoValue"))
            )

        if Replica_Rules:
            ReplicationConfiguration = s3.ReplicationConfiguration(
                "", Role=GetAtt(f"Role{resname}Replica", "Arn"), Rules=Replica_Rules
            )
            r_Bucket.ReplicationConfiguration = If(
                f"{resname}Replica", ReplicationConfiguration, Ref("AWS::NoValue")
            )

        PolicyStatementReplicaResources = []
        for m, w in v["PolicyStatementReplica"]["Resource"].items():
            polstatname = f"{resname}PolicyStatementReplicaResource{m}"
            # conditions
            add_obj(get_condition(f"{polstatname}Prefix", "not_equals", "none"))

            PolicyStatementReplicaResources.append(
                If(
                    f"{polstatname}Prefix",
                    get_subvalue(
                        "arn:aws:s3:::%s/${1M}*" % bucket_name, f"{polstatname}Prefix"
                    ),
                    Ref("AWS::NoValue"),
                )
            )

        BucketPolicyStatement = []

        r_Policy = S3BucketPolicy(
            f"BucketPolicy{name}",
            key=v,
            Condition=resname,
            Bucket=Ref(resname),
        )
        r_Policy.PolicyDocument["Statement"] = BucketPolicyStatement

        # At least one statement must be always present,
        # create a simple one with no conditions
        BucketPolicyStatement.extend(S3BucketPolicyStatementBase(resname))

        BucketPolicyStatement.extend(
            S3BucketPolicyStatementReplica(resname, PolicyStatementReplicaResources)
        )

        BucketPolicyStatement.extend(
            S3BucketPolicyStatementRead(resname, PolicyReadPrincipal)
        )

        BucketPolicyStatement.extend(
            S3BucketPolicyStatementWrite(resname, PolicyWritePrincipal)
        )

        r_IAMPolicyReplica = IAMPolicyBucketReplica(
            f"IAMPolicyReplicaBucket{name}",
            bucket=resname,
            bucket_name=bucket_name,
            mapname=f"{resname}ReplicationConfigurationRules",
            key=v["Replication"]["ConfigurationRules"],
        )

        try:
            bucket_policies = getattr(cfg, "BucketPolicy")
        except Exception:
            pass
        else:
            for policy_name, policy_value in bucket_policies.items():
                BucketPolicyStatement.append(get_dictvalue(policy_value))

        if "WebsiteConfiguration" in v:
            r_Bucket.WebsiteConfiguration = s3.WebsiteConfiguration(
                f"{resname}WebsiteConfiguration"
            )
            auto_get_props(r_Bucket.WebsiteConfiguration)

        if "PolicyStatement" in v:
            FixedStatements = []
            for fsn, fsv in v["PolicyStatement"].items():
                FixedStatement = IAMPolicyStatement(fsv)
                FixedStatement["Principal"] = {"AWS": eval(fsv["Principal"])}
                FixedStatement["Sid"] = fsv["Sid"]
                FixedStatements.append(FixedStatement)
            BucketPolicyStatement.extend(FixedStatements)

        if "PolicyStatementExGetObjectPrincipal" in v:
            BucketPolicyStatement.extend(
                S3BucketPolicyStatementAllowGetObject(
                    resname,
                    get_endvalue(f"{resname}PolicyStatementExGetObjectPrincipal"),
                    "AllowGetObjectExPrincipal",
                )
            )

        PolicyCloudFrontOriginAccessIdentityPrincipal = []
        if "CloudFrontOriginAccessIdentity" in v:
            identityname = v["CloudFrontOriginAccessIdentity"]
            identityresname = f"CloudFrontOriginAccessIdentity{identityname}"

            PolicyCloudFrontOriginAccessIdentityPrincipal.append(
                Sub(
                    "arn:aws:iam::cloudfront:user/"
                    "CloudFront Origin Access Identity ${%s}" % identityresname
                )
            )

            for ixn, ixv in v["CloudFrontOriginAccessIdentityExtra"].items():
                ixname = f"{resname}CloudFrontOriginAccessIdentityExtra{ixn}"
                # conditions
                add_obj(get_condition(ixname, "not_equals", "none"))

                PolicyCloudFrontOriginAccessIdentityPrincipal.append(
                    If(
                        ixname,
                        get_subvalue(
                            "arn:aws:iam::cloudfront:user/"
                            "CloudFront Origin Access Identity ${1M}",
                            ixname,
                        ),
                        Ref("AWS::NoValue"),
                    )
                )

            # conditions
            identitycondname = f"{resname}CloudFrontOriginAccessIdentity"
            c_identity = {
                identitycondname: And(
                    Condition(resname),
                    get_condition("", "not_equals", "none", identitycondname),
                )
            }

            add_obj(c_identity)

            # resources
            BucketPolicyStatement.extend(
                S3BucketPolicyStatementAllowGetObject(
                    resname,
                    PolicyCloudFrontOriginAccessIdentityPrincipal,
                    "AllowCFAccess",
                )
            )

            r_OriginAccessIdentity = CFOriginAccessIdentity(
                identityresname, comment=identityname, Condition=identitycondname
            )

            add_obj(r_OriginAccessIdentity)

            # outputs
            o_OriginAccessIdentity = Output(
                identityresname, Value=Ref(identityresname), Condition=identitycondname
            )

            add_obj(o_OriginAccessIdentity)

        add_obj([r_Bucket, r_Policy, r_IAMPolicyReplica])

        # outputs
        o_Bucket = Output(resname)
        outvalue = If(resname, Ref(resname), Sub(bucket_name))
        if "OutputValueRegion" in v:
            condname = f"{resname}OutputValueRegion"
            # conditions
            add_obj(get_condition(condname, "not_equals", "AWSRegion"))

            outvalue = If(
                condname,
                Sub(
                    "${Region}-%s" % bucket_name.replace("${AWS::Region}-", "", 1),
                    **{"Region": get_endvalue(condname)},
                ),
                outvalue,
            )
            o_Bucket.Export = Export(resname)
        else:
            o_Bucket.Condition = resname

        o_Bucket.Value = outvalue

        add_obj([o_Bucket])



import troposphere.events as eve

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


def EVE_EventRules(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"

        ibox_lo_cfg = v.get("IBOX_LINKED_OBJ", {})
        if isinstance(ibox_lo_cfg, str):
            ibox_lo_cfg = {"Base": ibox_lo_cfg}

        # resources
        Targets = []
        for m, w in v["Targets"].items():
            targetname = f"{resname}Targets{m}"
            Target = eve.Target(targetname)

            if m.startswith("Lambda"):
                permname = "%s%s" % (m.replace("Lambda", "LambdaPermission"), resname)
                # create ad hoc IBOX_LINKED_OBJ
                lo_cfg = {
                    "Key": "LambdaPermission",
                    "Type": "EventsRule",
                    "Name": f"LambdaPermission{m}{resname}",
                    "Conf": {
                        "IBOX_RESNAME": permname,
                        "IBOX_LINKED_OBJ_NAME": w["Arn"],
                        "IBOX_LINKED_OBJ_INDEX": f"GetAtt('{resname}', 'Arn')",
                    },
                }
                ibox_lo_cfg[m] = lo_cfg
            if m.startswith("ECSCluster"):
                # add common "fixed" props
                auto_get_props(
                    Target, mapname="EventsRuleTargetECSCluster", indexname=m
                )

            # add props found in yaml cfg
            auto_get_props(Target)
            Targets.append(Target)

        getattr(cfg, resname).update({"IBOX_LINKED_OBJ": ibox_lo_cfg})
        r_Rule = eve.Rule(resname)
        auto_get_props(r_Rule, indexname=n)
        r_Rule.Targets = Targets

        add_obj(r_Rule)



import troposphere.ecs as ecs

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)
from .ec2 import SecurityGroupEcsService, SecurityGroupRuleEcsService


def ECS_ContainerDefinition():
    Containers = []
    for n, v in cfg.ContainerDefinitions.items():
        name = f"ContainerDefinitions{n}"  # Ex. ContainerDefinitions1

        EnvValue_Out_String = []
        EnvValue_Out_Map = {}
        for m, w in v["Environment"].items():
            if m.startswith("Env"):
                continue
            envname = f"{name}Environment{m}"
            envkeyname = w["Name"]
            # parameters
            p_EnvValue = Parameter(
                f"{envname}Value",
                Description=f"{envkeyname} - empty for default based on env/role",
            )

            # If key NoParam is present skip adding Parameters
            # (usefull as they have a limited max number)
            if "NoParam" not in w:
                add_obj(p_EnvValue)

            EnvValue_Out_String.append("%s=${%s}" % (envkeyname, envkeyname))
            EnvValue_Out_Map.update({envkeyname: get_endvalue(f"{envname}Value")})

        # resources
        Container = ecs.ContainerDefinition(name)
        auto_get_props(Container, indexname=n)

        if len(cfg.ContainerDefinitions) == 1:
            # parameters
            p_UseTaskCpu = Parameter(
                f"{name}UseTaskCpu",
                Description="Empty for mapped value - Use Task Cpu Value, if present, for Container Cpu",
                AllowedValues=["", "true", "false"],
            )

            add_obj(p_UseTaskCpu)

            # conditions
            c_UseTaskCpu = {
                f"{name}UseTaskCpu": And(
                    Condition("CpuTask"),
                    get_condition("", "equals", "true", f"{name}UseTaskCpu"),
                )
            }

            add_obj(c_UseTaskCpu)

            Container.Cpu = If(
                f"{name}UseTaskCpu", get_endvalue("Cpu"), get_endvalue(f"{name}Cpu")
            )
            Container.Memory = If(
                "LaunchTypeFarGate",
                get_endvalue("Memory"),
                get_endvalue(f"{name}Memory"),
            )

        Containers.append(Container)

        # outputs
        o_EnvValueOut = Output(
            f"{name}Environment",
            Value=Sub(",".join(EnvValue_Out_String), **EnvValue_Out_Map),
        )

        add_obj(o_EnvValueOut)

    return Containers


def ECS_TaskDefinition(key):
    for n, v in getattr(cfg, key).items():
        if not v.get("IBOX_ENABLED", True):
            continue
        mapname = f"{key}{n}"

        # Resources
        R_TaskDefinition = ecs.TaskDefinition(mapname)
        auto_get_props(R_TaskDefinition)
        R_TaskDefinition.ContainerDefinitions = ECS_ContainerDefinition()

        add_obj([R_TaskDefinition])


def ECS_Service(key):
    # Resources
    R_SG = SecurityGroupEcsService("SecurityGroupEcsService")
    if "External" in cfg.LoadBalancer:
        SGRule = SecurityGroupRuleEcsService(scheme="External")
        R_SG.SecurityGroupIngress.append(SGRule)

    if "Internal" in cfg.LoadBalancer:
        SGRule = SecurityGroupRuleEcsService(scheme="Internal")
        R_SG.SecurityGroupIngress.append(SGRule)

    add_obj(R_SG)

    for n, v in getattr(cfg, key).items():
        if not v["IBOX_ENABLED"]:
            continue
        mapname = f"{key}{n}"

        # delete not used LoadBalancers configuration, so that auto_get_props
        # do not find it
        for m in ["External", "Internal"]:
            if m not in cfg.LoadBalancer:
                del v["LoadBalancers"][m]
        if not v["LoadBalancers"]:
            # delete if empty to maintain compatibility with previous conf
            del v["LoadBalancers"]

        r_Service = ecs.Service(mapname)
        auto_get_props(r_Service)

        if cfg.LoadBalancer:
            r_Service.Role = If(
                "NetworkModeAwsVpc", Ref("AWS::NoValue"), get_expvalue("RoleECSService")
            )

        # When creating a service that specifies multiple target groups,
        # the Amazon ECS service-linked role must be created.
        # The role is created by omitting the Role property
        # in AWS CloudFormation
        if all(k in cfg.LoadBalancer for k in ["External", "Internal"]):
            r_Service.Role = Ref("AWS::NoValue")

        add_obj(r_Service)
